#include <QPainter>
#include <QDebug>

#include "edge.hh"
#include "node.hh"

#include <math.h>
#include <cassert>

static const double Pi = M_PI;
static double TwoPi = 2.0 * Pi;

//! [0]
edge::edge(node *sourcenode, node *destnode, const QString& reason)
    : arrowSize(10)
{
    setAcceptedMouseButtons(0);
    source = sourcenode;
    dest = destnode;
    source->add_edge(this);
    dest->add_edge(this);
	reasons.push_back(reason);
    adjust();
	adjust_reason();
}

void edge::adjust_reason()
{
	reason_txt = reasons.join("\n");

	QGraphicsSimpleTextItem text;
	text.setText(reason_txt);
	bounding_text = text.boundingRect();
}

edge::~edge()
{
	source->remove_edge(this);
	dest->remove_edge(this);
}

node *edge::sourcenode() const
{
    return source;
}

node *edge::destnode() const
{
    return dest;
}

void edge::add_reason(const QString& reason) 
{
	reasons.push_back(reason);
	adjust_reason();
}

void edge::remove_reason(const QString& reason) 
{
	reasons.removeOne(reason);
	adjust_reason();
}

bool intersect(const QLineF& line, const QRectF& rect, QPointF* intersectionPoint,
			   const QPointF& src, const QPointF& dst)
{
	QLineF topLine(rect.topLeft(), rect.topRight());
	QLineF leftLine(rect.topLeft(), rect.bottomLeft());
	QLineF rightLine(rect.topRight(), rect.bottomRight());
	QLineF bottomLine(rect.bottomLeft(), rect.bottomRight());

	if (line.intersect(topLine, intersectionPoint) == QLineF::BoundedIntersection) {
		if (src.x() > dst.x()) 
			intersectionPoint->setX(intersectionPoint->x() - 5);
		else
			intersectionPoint->setX(intersectionPoint->x() + 5);
		return true;
	}
	if (line.intersect(leftLine, intersectionPoint) == QLineF::BoundedIntersection) {
		if (src.y() > dst.y()) 
			intersectionPoint->setY(intersectionPoint->y() + 5);
		else
			intersectionPoint->setY(intersectionPoint->y() - 5);
		return true;
	}
	if (line.intersect(rightLine, intersectionPoint) == QLineF::BoundedIntersection) {
		if (src.y() > dst.y()) 
			intersectionPoint->setY(intersectionPoint->y() + 5);
		else
			intersectionPoint->setY(intersectionPoint->y() - 5);
		return true;
	}

	if (line.intersect(bottomLine, intersectionPoint) == QLineF::BoundedIntersection) {
		if (src.x() > dst.x()) 
			intersectionPoint->setX(intersectionPoint->x() - 5);
		else
			intersectionPoint->setX(intersectionPoint->x() + 5);
		return true;
	}

	return false;
}

void edge::adjust()
{
    if (!source || !dest)
        return;

	QRectF bounding_src = source->boundingRect();
	QRectF bounding_dst = dest->boundingRect();
	bounding_src.translate(source->pos());
	bounding_dst.translate(dest->pos());

	QLineF line(bounding_src.center(), bounding_dst.center());

    qreal length = line.length();

    prepareGeometryChange();

    if (length > qreal(20.)) {
		intersect(line, bounding_src, &sourcePoint, source->pos(), dest->pos());
		intersect(line, bounding_dst, &destPoint, source->pos(), dest->pos());
    } else {
        sourcePoint = destPoint = line.p1();
    }
}

QRectF edge::boundingRect() const
{
    if (!source || !dest)
        return QRectF();

    qreal penWidth = 1;
    qreal extra = (penWidth + arrowSize) / 2.0;

    return QRectF(sourcePoint, QSizeF(destPoint.x() - sourcePoint.x(),
                                      destPoint.y() - sourcePoint.y()))
        .normalized()
        .adjusted(-extra, -extra, extra, extra);
}

void edge::paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *)
{
    if (!source || !dest)
        return;

    QLineF line(sourcePoint, destPoint);

    if (qFuzzyCompare(line.length(), qreal(0.)))
        return;

    // Draw the line itself
    painter->setPen(QPen(Qt::black, 1, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
    painter->drawLine(line);

    // Draw the arrows
    double angle = ::acos(line.dx() / line.length());
    if (line.dy() >= 0)
        angle = TwoPi - angle;

    QPointF sourceArrowP1 = sourcePoint + QPointF(sin(angle + Pi / 3) * arrowSize,
                                                  cos(angle + Pi / 3) * arrowSize);
    QPointF sourceArrowP2 = sourcePoint + QPointF(sin(angle + Pi - Pi / 3) * arrowSize,
                                                  cos(angle + Pi - Pi / 3) * arrowSize);   
    QPointF destArrowP1 = destPoint + QPointF(sin(angle - Pi / 3) * arrowSize,
                                              cos(angle - Pi / 3) * arrowSize);
    QPointF destArrowP2 = destPoint + QPointF(sin(angle - Pi + Pi / 3) * arrowSize,
                                              cos(angle - Pi + Pi / 3) * arrowSize);

    painter->setBrush(Qt::black);
    painter->drawPolygon(QPolygonF() << line.p2() << destArrowP1 << destArrowP2);
	painter->save();
	
	if (source->pos().x() > dest->pos().x()) {
		if (source->pos().y() > dest->pos().y()) {
			angle-= Pi;
			painter->translate(destPoint);
		} else {
			angle-= Pi;
			painter->translate(destPoint +
					QPointF(-bounding_text.height() * sin(angle),
							-bounding_text.height() * cos(angle)));
		}

	} else {
		if (source->pos().y() > dest->pos().y()) {
			painter->translate(sourcePoint);
		} else {
			painter->translate(sourcePoint + 
					QPointF(-bounding_text.height() * sin(angle),
							-bounding_text.height() * cos(angle)));
		}
	}

	qreal angle_deg = (-180.0 * angle / Pi) + 360;
	painter->rotate(angle_deg);
	QRectF bounding = bounding_text;
	bounding.setWidth(line.length());
	painter->drawText(bounding, reason_txt, Qt::AlignHCenter|Qt::AlignVCenter);
	painter->restore();
}
