
#include <edge.hh>
#include <graph_widget.hh>
#include <node.hh>

#include <QtGui>

#include <math.h>


graph_widget::graph_widget(QWidget* parent):
	QGraphicsView(parent), timerId(0)
{
    QGraphicsScene *scene = new QGraphicsScene(this);
    scene->setItemIndexMethod(QGraphicsScene::NoIndex);
//    scene->setSceneRect(-200, -200, 400, 400);
    setScene(scene);
    setCacheMode(CacheBackground);
    setViewportUpdateMode(BoundingRectViewportUpdate);
    setRenderHint(QPainter::Antialiasing);
    setTransformationAnchor(AnchorUnderMouse);
    scale(qreal(0.8), qreal(0.8));
    setMinimumSize(400, 400);
    setWindowTitle(tr("Hyper Viewer"));
}

graph_widget::~graph_widget() {}

void graph_widget::wheelEvent(QWheelEvent *event)
{
    scaleView(pow((double)2, -event->delta() / 240.0));
}

void graph_widget::timerEvent(QTimerEvent *event)
{
    Q_UNUSED(event);

    QList<node *> nodes;
    foreach (QGraphicsItem *item, scene()->items()) {
        if (node *pnode = qgraphicsitem_cast<node *>(item))
            nodes << pnode;
    }

    foreach (node *pnode, nodes)
        pnode->calculateForces();

    bool itemsMoved = false;
    foreach (node *pnode, nodes) {
        if (pnode->advance())
            itemsMoved = true;
    }

    if (!itemsMoved) {
        killTimer(timerId);
        timerId = 0;
    }
}

void graph_widget::itemMoved()
{
    if (!timerId) {
        timerId = startTimer(1000 / 25);
	}
}

void graph_widget::scaleView(qreal scaleFactor)
{
    qreal factor = transform().scale(scaleFactor, scaleFactor).mapRect(QRectF(0, 0, 1, 1)).width();
    if (factor < 0.07 || factor > 100)
        return;

    scale(scaleFactor, scaleFactor);
}

void graph_widget::add_node(const QString& name)
{
	qDebug() << "adding node " << name;

	if (node_table.contains(name)) {
		qDebug() << "Node " << name << " already in the table ";
		return;
	}

    node *n = new node(name, this);
	node_table[name] = n;
	scene()->addItem(n);
	itemMoved();
}

void graph_widget::remove_node(const QString& name)
{
	qDebug() << "removing node " << name;

	QHash<QString, node*>::iterator it;
	it = node_table.find(name);
	if (it == node_table.end()) {
		qDebug() << "Node " << name << " not found in table ";
		return;
	}

	node *n = it.value();
	const QList<edge*>& list = n->edges();
    foreach (edge *e, list) {
		 QHash<QString, edge*>::iterator it1 = edge_key_table.begin();
		 while (it1 != edge_key_table.end()) {
			 if (it1.value() == e) {
				 reason_table.remove(it1.key());
				 it1 = edge_key_table.erase(it1);
			 } else {
				 ++it1;
			 }
		 }
		 QString src_dst = e->sourcenode()->name() + "_" + e->destnode()->name();
		 edge_table.remove(src_dst);
		 scene()->removeItem(e);
		 delete e;
	}

	scene()->removeItem(n);
	node_table.erase(it);
	delete n;
	itemMoved();
}

void graph_widget::add_connection(const QString& key,
								  const QString& src, const QString& dst,
								  const QString& reason)
{
	qDebug() << "Add connection " << reason << " between " << src << " and " << dst << " key : " << key;
	QHash<QString, edge*>::iterator it;
	it = edge_key_table.find(key);
	if (it != edge_key_table.end()) {
		qDebug() << "Critical error detected : non unique key used !!";
		return;
	}

	QSet<QString>::iterator it3;
	it3 = removed_connection.find(key);
	if (it3 != removed_connection.end()) {
		removed_connection.erase(it3);
		return;
	}

	QString src_dst = src + "_" + dst;
	it = edge_table.find(src_dst);
	if (it != edge_table.end()) {
		it.value()->add_reason(reason);
		edge_key_table[key] = it.value();
		reason_table[key] = reason;
		return;
	}

	QHash<QString, node*>::iterator it1, it2;
	it1 = node_table.find(src);
	it2 = node_table.find(dst);
	if (it1 == node_table.end() || it2 == node_table.end()) {
		qDebug() << "Can't find " << src << " or " << dst;
		return;
	}

	edge* e = new edge(it1.value(), it2.value(), reason);
	edge_table[src_dst] = e;
	edge_key_table[key] = e;
	reason_table[key] = reason;
	scene()->addItem(e);
	itemMoved();
}

void graph_widget::remove_connection(const QString& key)
{
	qDebug() << "Remove connection associated to key " << key;

	QHash<QString, edge*>::iterator it1;
	QHash<QString, QString>::iterator it2;

	it1 = edge_key_table.find(key);
	it2 = reason_table.find(key);
	if (it1 == edge_key_table.end() || it2 == reason_table.end()) {
		qDebug() << "Can't find " << key << " in the edge table";
		removed_connection.insert(key);
		return;
	}

	edge* e = it1.value();
	e->remove_reason(it2.value());
	reason_table.erase(it2);
	if (e->number_reason() == 0) {
		QString src_dst = e->sourcenode()->name() + "_" +
						  e->destnode()->name();
		edge_table.remove(src_dst);
		scene()->removeItem(e);
		delete e;
		itemMoved();
	}
	edge_key_table.erase(it1);
}
								     
