#include <vector>
#include <string>

#include <boost/date_time/posix_time/posix_time.hpp>

#include <network/msg.hh>
#include <network/ping.hh>
#include <network/nameserver.hh>
#include <network/server_tcp_impl.hh>

#include <model/discover_root.hh>

#include <hyper_live_node.hh>

#include <QRegExp>

namespace {
	typedef boost::mpl::vector<hyper::network::log_msg,
							   hyper::network::inform_new_agent,
							   hyper::network::inform_death_agent,
							   hyper::network::list_agents
							   > input_msg;
	typedef boost::mpl::vector<boost::mpl::void_> output_msg;

	typedef boost::make_variant_over<input_msg>::type input_variant;
	typedef boost::make_variant_over<output_msg>::type output_variant;

	struct logger_visitor : public boost::static_visitor<output_variant>
	{
		std::vector<hyper::network::log_msg> & log_msgs_;
		hyper_node* node;
		QRegExp constraint_rqst, value_rqst, answer_rqst;

		logger_visitor(std::vector<hyper::network::log_msg>& log_msgs, hyper_node* node) : 
			log_msgs_(log_msgs), node(node),
			constraint_rqst("\\[([a-zA-Z][a-zA-Z0-9_]*), (\\d+)\\] "
							"Handling (.*)"),
			value_rqst("\\[([a-zA-Z][a-zA-Z0-9_]*), (\\d+)\\] "
					   "Request for  the value of (.*)"),
			answer_rqst("\\[([a-zA-Z][a-zA-Z0-9_]*), (\\d+)\\] "
						"Final answer")
		{}

		output_variant operator() (const hyper::network::log_msg& msg) const
		{
			log_msgs_.push_back(msg);
			QString content = msg.msg.c_str();
			if (constraint_rqst.exactMatch(content)) {
				QString src = constraint_rqst.cap(1);
				QString id = constraint_rqst.cap(2);
				QString constraint = constraint_rqst.cap(3);
				constraint.replace("%", "");
				QString key = src + id;

				node->emit_add_link(key, src, msg.src.c_str(), constraint);
				
			} else if (value_rqst.exactMatch(content)) {
				QString src = value_rqst.cap(1);
				QString id = value_rqst.cap(2);
				QString value = value_rqst.cap(3);
				QString dst = msg.src.c_str();

				QString reason = dst + "::" + value;
				QString key = src + id;

				node->emit_add_link(key, dst, src, reason);
			} else if (answer_rqst.exactMatch(content)) {
				QString src = answer_rqst.cap(1);
				QString id = answer_rqst.cap(2);
				QString key = src + id;

				node->emit_remove_link(key);
			}


			return boost::mpl::void_();
		}

		output_variant operator() (const hyper::network::inform_death_agent& msg) const
		{
			std::vector<std::string>::const_iterator it;
			for (it = msg.dead_agents.begin(); it != msg.dead_agents.end(); ++it)
				node->emit_remove_agent(it->c_str());

			return boost::mpl::void_();
		}

		output_variant operator() (const hyper::network::inform_new_agent& msg) const
		{
			std::vector<std::string>::const_iterator it;
			for (it = msg.new_agents.begin(); it != msg.new_agents.end(); ++it)
				node->emit_new_agent(it->c_str());

			return boost::mpl::void_();
		}

		output_variant operator() (const hyper::network::list_agents& msg) const
		{
			std::vector<std::string>::const_iterator it;
			for (it = msg.all_agents.begin(); it != msg.all_agents.end(); ++it)
				node->emit_new_agent(it->c_str());

			return boost::mpl::void_();
		}
	};

	struct cmp_log_msg_by_date
	{
		bool operator() (const hyper::network::log_msg& msg1, 
						 const hyper::network::log_msg& msg2) const
		{
			return (msg1.date < msg2.date);
		}
	};

	struct too_young_msg
	{
		boost::posix_time::time_duration delay_;

		too_young_msg(boost::posix_time::time_duration delay) : delay_(delay) {}

		bool operator() (const hyper::network::log_msg msg) const
		{
			boost::posix_time::ptime oldest(boost::posix_time::microsec_clock::local_time());
			oldest -= delay_;
			return (msg.date < oldest);
		}
	};

	struct display_log 
	{
		void operator() (const hyper::network::log_msg& msg)
		{
			std::cout << "[" << boost::posix_time::to_iso_string(msg.date) << "]";
			std::cout << "[" << msg.src << "] ";
			std::cout << msg.msg << std::endl;
		}
	};

	struct periodic_check
	{
		boost::asio::io_service& io_s_;

		boost::posix_time::time_duration delay_;
		boost::asio::deadline_timer timer_;

		std::vector<hyper::network::log_msg> & log_msgs_;


		periodic_check(boost::asio::io_service& io_s, 
					   boost::posix_time::time_duration delay,
					   std::vector<hyper::network::log_msg>& log_msgs):
			io_s_(io_s), delay_(delay), timer_(io_s_),
			log_msgs_(log_msgs)
		{}

		void handle_timeout(const boost::system::error_code& e)
		{
			/* Make a chance to reorder msg in the right way */
			if (!e) {
				std::sort(log_msgs_.begin(), log_msgs_.end(), cmp_log_msg_by_date());
				std::vector<hyper::network::log_msg>::iterator it;
				it = std::find_if(log_msgs_.begin(), log_msgs_.end(), too_young_msg(delay_*2));
				if (it != log_msgs_.begin())
				{
					std::for_each(log_msgs_.begin(), it, display_log());
					log_msgs_.erase(log_msgs_.begin(), it); 
				}

				run();
			}
		}

		void run()
		{
			timer_.expires_from_now(delay_);
			timer_.async_wait(boost::bind(&periodic_check::handle_timeout, 
									  this,
									  boost::asio::placeholders::error));
		}
	};
}

class hyper_pimpl {
	private:
		typedef hyper::network::tcp::server<input_msg, output_msg, logger_visitor> logger_server;
		boost::asio::io_service io_s;
		std::vector<hyper::network::log_msg> msgs;
		hyper::model::discover_root discover;
		hyper::network::name_client name_client_;
		logger_visitor vis;
		logger_server serv;
		periodic_check check;
		hyper::network::ping_process ping;

		hyper_pimpl(const hyper_pimpl&);

	public:
		hyper_pimpl(hyper_node *node) :
			discover(),
			name_client_(io_s, discover.root_addr(), discover.root_port()),
			vis(msgs, node),
			serv(vis, io_s),
			check(io_s, boost::posix_time::milliseconds(100), msgs),
			ping(io_s, boost::posix_time::milliseconds(100),
					   "logger", discover.root_addr(), discover.root_port())
		{
			bool res = name_client_.register_name("logger", serv.local_endpoints());
			if (res == false) {
				std::cerr << "Failed to register logger, exiting ... " << std::endl;
				exit(EXIT_FAILURE); // XXX
			}
		}

		void run() { 
			ping.run();
			check.run();
			io_s.run();
		}
};

hyper_live_node::hyper_live_node() {
	impl = new hyper_pimpl(this);
}

hyper_live_node::~hyper_live_node() {
	delete impl;
}

void hyper_live_node::run() {
	impl->run();
}

