#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#include <hyper_replay_node.hh>

#include <QFile>
#include <QTextStream>
#include <QRegExp>

hyper_replay_node::hyper_replay_node(const std::string& s, double slowing_factor) :
	slowing_factor(slowing_factor), must_pause(false)
{
	QString line;
	QFile myFile(s.c_str());
	myFile.open(QIODevice::ReadOnly);
	QTextStream stream( &myFile );

	QRegExp constraint_rqst("\\[([0-9]*T[0-9]*\\.[0-9]*)\\]\\[([a-zA-Z][a-zA-Z0-9_]*)\\] "
							"\\[([a-zA-Z][a-zA-Z0-9_]*), (\\d+)\\] Handling (.*)");
	QRegExp value_rqst("\\[([0-9]*T[0-9]*\\.[0-9]*)\\]\\[([a-zA-Z][a-zA-Z0-9_]*)\\] "
					   "\\[([a-zA-Z][a-zA-Z0-9_]*), (\\d+)\\] "
					   "Request for  the value of (.*)");
	QRegExp answer_rqst("\\[([0-9]*T[0-9]*\\.[0-9]*)\\]\\[([a-zA-Z][a-zA-Z0-9_]*)\\] "
						"\\[([a-zA-Z][a-zA-Z0-9_]*), (\\d+)\\] "
						"Final answer");
	QRegExp new_agent("\\[([0-9]*T[0-9]*\\.[0-9]*)\\]\\[root\\] "
						"New agent\\(s\\) in the system : ([a-zA-Z][a-zA-Z0-9_]*),");

	QRegExp death_agent("\\[([0-9]*T[0-9]*\\.[0-9]*)\\]\\[([a-zA-Z][a-zA-Z0-9_]*)\\] "
						"Receive information about the death of agent\\(s\\) : ([a-zA-Z][a-zA-Z0-9_]*),");

	std::map<QString, int> map_index;

	line = stream.readLine();
	while (! line.isNull()) {
		if (constraint_rqst.exactMatch(line)) {
			details::add_link ev_(
				constraint_rqst.cap(3),
				constraint_rqst.cap(4),
				constraint_rqst.cap(2),
				constraint_rqst.cap(5)
			);

			details::event ev;
			ev.ev = ev_;
			ev.abs_time = boost::posix_time::from_iso_string(qPrintable(constraint_rqst.cap(1)));

			events.push_back(ev);
			
			map_index[ev_.src + ev_.id] = events.size() - 1;
		}

		if (value_rqst.exactMatch(line)) {
			details::add_link ev_(
				value_rqst.cap(3),
				value_rqst.cap(4),
				value_rqst.cap(2),
				value_rqst.cap(5));

			details::event ev;
			ev.ev = ev_;
			ev.abs_time = boost::posix_time::from_iso_string(qPrintable(value_rqst.cap(1)));

			events.push_back(ev);

			map_index[ev_.src + ev_.id] = events.size() - 1;
		}

		if (answer_rqst.exactMatch(line)) {
			details::remove_link ev_(
				answer_rqst.cap(3),
				answer_rqst.cap(4));

			ev_.reverse = map_index[ev_.src + ev_.id];

			details::event ev;
			ev.ev = ev_;
			ev.abs_time = boost::posix_time::from_iso_string(qPrintable(answer_rqst.cap(1)));

			events.push_back(ev);
		}

		if (new_agent.exactMatch(line)) {
			details::new_agent ev_(new_agent.cap(2));

			details::event ev;
			ev.ev = ev_;
			ev.abs_time = boost::posix_time::from_iso_string(qPrintable(new_agent.cap(1)));

			events.push_back(ev);
		}

		if (death_agent.exactMatch(line)) {
			details::remove_agent ev_(death_agent.cap(3));

			details::event ev;
			ev.ev = ev_;
			ev.abs_time = boost::posix_time::from_iso_string(qPrintable(death_agent.cap(1)));
			events.push_back(ev);
		}

		line = stream.readLine();
	}

	for (size_t i = 1; i < events.size(); i++)
	{
		events[i].rel_time = events[i].abs_time - events[i-1].abs_time;
	}
}

hyper_replay_node::~hyper_replay_node()
{}

struct send_event : public boost::static_visitor<void>
{
	hyper_node* node;

	send_event(hyper_node* node) : node(node) {}

	void operator() (const details::add_link& add) const
	{
		QString key = add.src + add.id;

		node->emit_add_link(key, add.src, add.dst, add.content);
	}

	void operator() (const details::remove_link& rem) const
	{
		QString key = rem.src + rem.id;
		node->emit_remove_link(key);
	}

	void operator() (const details::new_agent& ag) const
	{
		node->emit_new_agent(ag.agent);
	}

	void operator() (const details::remove_agent& ag) const
	{
		node->emit_remove_agent(ag.agent);
	}
};

struct send_inverse_event : public boost::static_visitor<void>
{
	hyper_replay_node* node;

	send_inverse_event(hyper_replay_node* node): node(node) {}

	void operator() (const details::add_link& add) const
	{
		QString key = add.src + add.id;

		node->emit_remove_link(key);
	}

	void operator() (const details::remove_link& rem) const
	{
		boost::apply_visitor(send_event(node), node->events[rem.reverse].ev);
	}

	void operator() (const details::new_agent& ag) const
	{
		node->emit_remove_agent(ag.agent);
	}

	void operator() (const details::remove_agent& ag) const
	{
		node->emit_new_agent(ag.agent);
	}
};


void hyper_replay_node::run()
{
	for (current_idx = 0; current_idx < size() ; ++current_idx) {
		mutex.lock();
		while (must_pause) {
			waitRestart.wait(&mutex);
		}
		mutex.unlock();

		boost::apply_visitor(send_event(this), events[current_idx].ev);
		usleep(events[current_idx+1].rel_time.total_microseconds() * slowing_factor);
	}
}

void hyper_replay_node::pause()
{
	mutex.lock();
	must_pause = true;
	mutex.unlock();
}

void hyper_replay_node::restart()
{
	mutex.lock();
	must_pause = false;
	waitRestart.wakeAll();
	mutex.unlock();
}

void hyper_replay_node::move(int i)
{
	size_t idx(i);
	if (idx >= size()) idx = size() - 1;
	if (idx >= current_idx) {
		for (current_idx; current_idx <= idx; ++current_idx) {
			boost::apply_visitor(send_event(this), events[current_idx].ev);
		}
	} else {
		for (current_idx; current_idx > idx; --current_idx)
			boost::apply_visitor(send_inverse_event(this), events[current_idx].ev);
	}
}
