#include <QGraphicsScene>
#include <QGraphicsSceneMouseEvent>
#include <QGraphicsSimpleTextItem>
#include <QPainter>
#include <QStyleOption>
#include <QtGlobal>
#include <QDebug>

#include "node.hh"
#include "edge.hh"
#include "graph_widget.hh"

#include <math.h>

node::node(const QString& name, graph_widget *graphWidget)
    : graph(graphWidget), name_(name)
{
    setFlag(ItemIsMovable);
    setFlag(ItemSendsGeometryChanges);
    setCacheMode(DeviceCoordinateCache);
    setZValue(-1);

	QGraphicsSimpleTextItem text;
	text.setText(name_);
	font = text.font();
	font.setPointSize(20);
	font.setBold(true);
	font.setCapitalization(QFont::SmallCaps);
	text.setFont(font);
	QRectF bounding = text.boundingRect();
	width = bounding.width() + 30;
	height = bounding.height() + 15;
}

void node::add_edge(edge* e)
{
	edgeList.push_back(e);
	e->adjust();
}

void node::remove_edge(edge* e)
{
	edgeList.removeAll(e);
}

const QList<edge*>& node::edges() const
{
	return edgeList;
}

void node::calculateForces()
{
    if (!scene() || scene()->mouseGrabberItem() == this) {
        newPos = pos();
        return;
    }

    // Sum up all forces pushing this item away
    qreal xvel = 0;
    qreal yvel = 0;
    foreach (QGraphicsItem *item, scene()->items()) {
        node *pnode = qgraphicsitem_cast<node *>(item);
        if (!pnode || pnode == this)
            continue;

        QPointF vec = mapToItem(pnode, 0, 0);
        qreal dx = vec.x();
        qreal dy = vec.y();
        double l = 2.0 * (dx * dx + dy * dy);
        if (l > 0.1 && l < 100000.0) {
            xvel += (dx * 150.0) / l;
            yvel += (dy * 150.0) / l;
		} else if (l < 0.1) {
			/* If they are really near from another node, or "on it", start to
			 * generate a move */
			xvel += qrand() / ((double)RAND_MAX + 1);
			yvel += qrand() / ((double)RAND_MAX + 1);
		}
    }

    // Now subtract all forces pulling items together
    double weight = (edgeList.size() + 1) * 10;
    foreach (edge *e, edgeList) {
        QPointF vec;
        if (e->sourcenode() == this)
            vec = mapToItem(e->destnode(), 0, 0);
        else
            vec = mapToItem(e->sourcenode(), 0, 0);
		qreal l = sqrt(vec.x() * vec.x() + vec.y() * vec.y());
		if (l >= e->min_size() * 10.0) {
			xvel -= vec.x() / weight;
			yvel -= vec.y() / weight;
		}
    }

    if (qAbs(xvel) < 0.1 && qAbs(yvel) < 0.1)
        xvel = yvel = 0;

    QRectF sceneRect = scene()->sceneRect();
    newPos = pos() + QPointF(xvel, yvel);
    newPos.setX(qMin(qMax(newPos.x(), sceneRect.left() + 10), sceneRect.right() - 10));
    newPos.setY(qMin(qMax(newPos.y(), sceneRect.top() + 10), sceneRect.bottom() - 10));
}

bool node::advance()
{
    if (newPos == pos())
        return false;

    setPos(newPos);
    return true;
}

QRectF node::boundingRect() const
{
    qreal adjust = 2;
    return QRectF(-10 - adjust, -10 - adjust,
                  width + 3  + adjust, 
				  height + 3 + adjust);
}

QPainterPath node::shape() const
{
    QPainterPath path;
    path.addEllipse(-10, -10, width, height);
    return path;
}

void node::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *)
{

    painter->setPen(Qt::NoPen);
    painter->setBrush(Qt::darkGray);
    painter->drawEllipse(-7, -7, width, height);

    painter->setPen(QPen(Qt::black, 0));
    painter->setBrush(Qt::yellow);
	QRectF rect(-10, -10, width, height);
    painter->drawEllipse(rect);
	painter->setFont(font);
	painter->drawText(rect, name_, Qt::AlignHCenter|Qt::AlignVCenter);
}

QVariant node::itemChange(GraphicsItemChange change, const QVariant &value)
{
    switch (change) {
    case ItemPositionHasChanged:
        foreach (edge *e, edgeList)
            e->adjust();
        graph->itemMoved();
        break;
    default:
        break;
    };

    return QGraphicsItem::itemChange(change, value);
}

void node::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    update();
    QGraphicsItem::mousePressEvent(event);
}

void node::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
    update();
    QGraphicsItem::mouseReleaseEvent(event);
}
